from abc import ABC, abstractmethod
import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection
from mpl_toolkits.mplot3d.art3d import Line3DCollection
import numpy as np


class Plotter(ABC):
    def __init__(self, vertices, edges):
        """
        Constructor for Plotter class: collection of methods to easily plot domain and solution
        :type vertices: np.ndarray
        :type edges: np.ndarray
        """
        self.vertices = vertices
        # Save edges without self loops
        tmp = np.array(edges)
        mask = tmp[:, 0] != tmp[:, 1]
        self.edges = tmp[mask]
        self.loops = tmp[~mask]
        # Bounding box
        self.x_min, self.x_max = min(self.vertices[:, 0]), max(self.vertices[:, 0])
        self.y_min, self.y_max = min(self.vertices[:, 1]), max(self.vertices[:, 1])
        # Minimum space for self loops
        self.x_min -= 0.12
        self.x_max += 0.12
        self.y_min -= 0.12
        self.y_max += 0.12

    def _finalize_figure(self, fig, ax, title, interactive, show_axis, collection=None, legend_label=""):
        """
        Perform the final configuration needed to display the figure
        """
        # Set labels
        ax.set_xlabel("x")
        ax.set_ylabel("y")
        # Set limits
        ax.set_xlim(self.x_min, self.x_max)
        ax.set_ylim(self.y_min, self.y_max)
        # Title
        plt.title(title)
        # Add color bar
        if collection is not None:
            cbar = fig.colorbar(collection, ax=ax, pad=0.1)
            cbar.set_label(legend_label)
        # Show plot
        if interactive:
            plt.ion()
        else:
            plt.ioff()
        # Show axis
        if show_axis:
            plt.axis('on')
        else:
            plt.axis('off')
        plt.show()

    # Abstract methods

    @abstractmethod
    def _initialize_figure(self):
        pass

    @abstractmethod
    def _plot_arrows(self, ax, color, linewidths, vertices_label):
        pass

    @abstractmethod
    def _plot_vertices(self, ax, vert_label=False, **kwargs):
        pass

    @abstractmethod
    def _plot_edges(self, ax, **kwargs):
        pass

    @abstractmethod
    def _plot_self_loops(self, ax, color, linewidths, vertices_label):
        pass

    # Public interface

    def plot_graph(self, color="blue", directed=False, interactive=False, line_widths=2, show_axis=False, title="",
                   vertices_label=True):
        """
        Plot image with graph representation.
        :param color: color to use for plotting the edges
        :param directed: if True, plot the graph using arrows
        :param interactive: if True, display the plot in interactive mode
        :param line_widths: width of the lines
        :param show_axis: if True, display the axis of the figure
        :param title: title of the plot
        :param vertices_label: if True, label the vertices of the graph
        """
        # Initialize figure
        fig, ax = self._initialize_figure()
        # Plot edges
        if directed:
            self._plot_arrows(ax, color, line_widths, vertices_label)
        else:
            self._plot_edges(ax, color=color, linewidths=line_widths)
        # Plot self loops
        self._plot_self_loops(ax, color, line_widths, vertices_label)
        # Plot vertices
        self._plot_vertices(ax, vertices_label, c=color, alpha=1)
        # Display figure
        self._finalize_figure(fig, ax, title, interactive, show_axis)

    def plot_edges_weight(self, quantity, edge_width=3, interactive=False, legend_label="", max_value=None,
                          min_value=None, palette='viridis', show_axis=False, title=""):
        """
        Plot edges weight
        :param quantity: vector of weights to plot on edges
        :type quantity: np.ndarray
        :param edge_width: width to use for the edges
        :param interactive: if True, display the plot in interactive mode
        :param legend_label: label of the colormap
        :param max_value: maximum value to consider for the colormap, if None it's the actual maximum
        :type max_value: float
        :param min_value: minimum value to consider for the colormap, if None it's the actual minimum
        :type min_value: float
        :param palette: colormap palette
        :param show_axis: if True, display the axis of the figure
        :param title: title of the plot
        """
        if min_value is None:
            min_value = np.min(quantity)
        if max_value is None:
            max_value = np.max(quantity)
        # Initialize figure
        fig, ax = self._initialize_figure()
        # Add edges
        lc = self._plot_edges(ax, cmap=palette, linewidths=edge_width, array=quantity)
        lc.set_clim(vmin=min_value, vmax=max_value)
        # Display figure
        self._finalize_figure(fig, ax, title, interactive, show_axis, lc, legend_label)

    def plot_vertices_weight(self, quantity, interactive=False, legend_label="", max_value=None, min_value=None,
                             palette='viridis', show_axis=False, title="", vertices_size=100):
        """
        Plot vertices weight
        :param quantity: vector of weights to plot on edges
        :type quantity: np.ndarray
        :param interactive: if True, display the plot in interactive mode
        :param legend_label: label of the colormap
        :param max_value: maximum value to consider for the colormap, if None it's the actual maximum
        :type max_value: float
        :param min_value: minimum value to consider for the colormap, if None it's the actual minimum
        :type min_value: float
        :param palette: colormap palette
        :param show_axis: if True, display the axis of the figure
        :param title: title of the plot
        :param vertices_size: size of the vertices in the plot
        """
        if min_value is None:
            min_value = np.min(quantity)
        if max_value is None:
            max_value = np.max(quantity)
        # Initialize figure
        fig, ax = self._initialize_figure()
        # Plot edges
        self._plot_edges(ax, color='dimgray', linewidths=0.5)
        # Plot points
        points = self._plot_vertices(ax, c=quantity, cmap=palette, alpha=1, vmin=min_value, vmax=max_value,
                                     s=vertices_size, zorder=2)
        # Display figure
        self._finalize_figure(fig, ax, title, interactive, show_axis, points, legend_label)


class Plotter2d(Plotter):
    def __init__(self, vertices, edges):
        super().__init__(vertices, edges)

    def _initialize_figure(self):
        """
        Initialize the figure in 2D
        """
        return plt.subplots()

    def _plot_arrows(self, ax, color, linewidths, vertices_label):
        """
        Plot directed edges as arrows (2D)
        """
        # Distance coefficient
        alpha = 0.08 if vertices_label else 0.01
        # Get values
        x0, y0, dx, dy = [], [], [], []
        for e in self.edges:
            x_s, y_s = self.vertices[e[0]]
            x_t, y_t = self.vertices[e[1]]
            d1, d2 = x_t - x_s, y_t - y_s
            x0.append(x_s + d1 * alpha)
            y0.append(y_s + d2 * alpha)
            dx.append(d1 * (1 - 2 * alpha))
            dy.append(d2 * (1 - 2 * alpha))
        # Plot arrows
        ax.quiver(x0, y0, dx, dy, angles='xy', scale_units='xy', scale=1, width=0.0015 * linewidths,
                  headwidth=4, headlength=4, color=color)

    def _plot_vertices(self, ax, vert_label=False, **kwargs):
        """
        Plot vertices in 2D
        """
        if vert_label:
            ret = ax.scatter(self.vertices[:, 0], self.vertices[:, 1], s=600, zorder=2, facecolors='white',
                             edgecolors=kwargs['c'])
            for i, v in enumerate(self.vertices):
                ax.text(v[0], v[1], i, fontsize=12, color=kwargs['c'], ha='center', va='center')
        else:
            ret = ax.scatter(self.vertices[:, 0], self.vertices[:, 1], **kwargs)
        return ret

    def _plot_edges(self, ax, **kwargs):
        """
        Plot edges in an efficient way using LineCollection (2D)
        """
        # Extract line segments
        xs, ys = [], []
        for e in self.edges:
            x0, y0 = self.vertices[e[0]]
            x1, y1 = self.vertices[e[1]]
            xs.append([x0, x1])
            ys.append([y0, y1])
        segments = [np.column_stack([x, y]) for x, y in zip(xs, ys)]
        # Create LineCollection (2D)
        line_segments = LineCollection(segments, **kwargs)
        # Add to axes
        ax.add_collection(line_segments)
        return line_segments

    def _plot_self_loops(self, ax, color, linewidths, vertices_label):
        """
        Draw self loops in (2D).
        """
        # Circle radius
        radius = 0.05
        for e in self.loops:
            x0, y0 = self.vertices[e[0]]
            # Compute circle
            theta = np.linspace(0, np.pi * (1.7 if vertices_label else 1.8), 50)
            x = x0 + radius * np.cos(theta) - radius
            y = y0 + radius * np.sin(theta)
            # Plot circle
            ax.plot(x, y, color=color, lw=linewidths)
            # Final points as arrow
            ax.annotate("", xy=(x[-1], y[-1]), xytext=(x[-2], y[-2]),
                        arrowprops=dict(arrowstyle="->", color=color, lw=1.5 * linewidths))


class Plotter3d(Plotter):
    def __init__(self, vertices, edges):
        super().__init__(vertices, edges)
        # Bounding box (adding z coordinate)
        self.z_min, self.z_max = min(self.vertices[:, 2]), max(self.vertices[:, 2])
        # Minimum space for self loops
        self.z_min -= 0.12
        self.z_max += 0.12

    def _finalize_figure(self, fig, ax, title, interactive, show_axis, collection=None, legend_label=""):
        """
        Perform the final configuration needed to display the figure in 3d (add zlabel and zlim)
        """
        # Add z label and limits
        ax.set_zlabel("z")
        ax.set_zlim(self.z_min, self.z_max)
        super()._finalize_figure(fig, ax, title, interactive, show_axis, collection, legend_label)

    def _initialize_figure(self):
        """
        Initialize the figure in 3D
        """
        fig = plt.figure()
        ax = fig.add_subplot(projection='3d')
        return fig, ax

    def _plot_arrows(self, ax, color, linewidths, vertices_label):
        """
        Plot directed edges as arrows (3D)
        """
        # Distance coefficient
        alpha = 0.1 if vertices_label else 0.01
        # Get values
        x0, y0, z0, dx, dy, dz = [], [], [], [], [], []
        for e in self.edges:
            x_s, y_s, z_s = self.vertices[e[0]]
            x_t, y_t, z_t = self.vertices[e[1]]
            d1, d2, d3 = x_t - x_s, y_t - y_s, z_t - z_s
            x0.append(x_s + d1 * alpha)
            y0.append(y_s + d2 * alpha)
            z0.append(z_s + d3 * alpha)
            dx.append(d1 * (1 - 2 * alpha))
            dy.append(d2 * (1 - 2 * alpha))
            dz.append(d3 * (1 - 2 * alpha))
        # Plot arrows
        ax.quiver(x0, y0, z0, dx, dy, dz, arrow_length_ratio=0.05, length=1,
                  color=color, linewidth=linewidths)

    def _plot_vertices(self, ax, vert_label=False, **kwargs):
        """
        Plot vertices in 3D
        """
        if vert_label:
            ret = ax.scatter(self.vertices[:, 0], self.vertices[:, 1], self.vertices[:, 2], s=500, zorder=2, alpha=1,
                             facecolors='white', edgecolors=kwargs['c'])
            for i, v in enumerate(self.vertices):
                ax.text(v[0], v[1], v[2], i, fontsize=12, color=kwargs['c'], zorder=1e6, ha='center', va='center')
        else:
            ret = ax.scatter(self.vertices[:, 0], self.vertices[:, 1], self.vertices[:, 2], **kwargs)
        return ret

    def _plot_edges(self, ax, **kwargs):
        """
        Plot edges in an efficient way, with the possibility of creating a colormap
        """
        radius = 0.05
        # Extract line segments
        xs, ys, zs = [], [], []
        for e in self.edges:
            x0, y0, z0 = self.vertices[e[0]]
            x1, y1, z1 = self.vertices[e[1]]
            xs.append([x0, x1])
            ys.append([y0, y1])
            zs.append([z0, z1])
        segments = [np.column_stack([x, y, z]) for x, y, z in zip(xs, ys, zs)]
        # Create LineCollection (3D)
        line_segments = Line3DCollection(segments, **kwargs)
        # Add collection
        ax.add_collection(line_segments)
        # Return collection
        return line_segments

    def _plot_self_loops(self, ax, color, linewidths, vertices_label):
        """
        Draw self loops in (3D).
        """
        # Circle radius
        radius = 0.1
        for e in self.loops:
            x0, y0, z0 = self.vertices[e[0]]
            # Compute circle
            theta = np.linspace(0, np.pi * (1.7 if vertices_label else 1.8), 50)
            x = x0 + radius * np.cos(theta) - radius
            y = y0 + radius * np.sin(theta)
            z = np.full_like(theta, z0)
            # Plot circle
            ax.plot(x, y, z, color=color, lw=linewidths)
            # Final arrow with quiver
            xs, ys, zs = x[-2], y[-2], z[-2]
            xt, yt, zt = x[-1], y[-1], z[-1]
            ax.quiver(xs, ys, zs, xt - xs, yt - ys, zt - zs, arrow_length_ratio=2, color=color, linewidth=linewidths)



import numpy as np
import vtk

from .graph import MetricGraph


class WeightedGraph(MetricGraph):
    """
    Extend MetricGraph adding weights on vertices and edges

    :param vertices: vertices coordinates of the graph
    :type vertices: list | np.ndarray
    :param edges: connectivity matrix of the graph
    :type edges: list | np.ndarray
    """

    def __init__(self, vertices, edges):
        super().__init__(vertices, edges)
        self.__weights = dict()
        self.__types = dict()

    def __set_weight(self, name, value, nodal_weight):
        value = np.array(value).ravel()
        length = self.vertices.shape[0] if nodal_weight else self.edges.shape[0]
        if value.shape[0] != length:
            print(f"Warning: value shape should be ({length}, ) ")
            return
        self.__weights[name] = np.array(value)
        self.__types[name] = nodal_weight

    def set_edges_weight(self, name, value):
        """
        Add a weight on the edges

        :param name: name of the weight
        :type name: str
        :param value: value of the weight
        :type value: list | np.ndarray
        """
        self.__set_weight(name, value, False)

    def set_vertices_weight(self, name, value):
        """
        Add a weight on the vertices

        :param name: name of the weight
        :type name: str
        :param value: value of the weight
        :type value: list | np.ndarray
        """
        self.__set_weight(name, value, True)

    def get_weight(self, name):
        """
        Get a value that represent a weight, given its name

        :param name: name of the weight
        :type name: str
        :return: value of the weight
        :rtype: np.ndarray
        """
        return self.__weights.get(name)

    def get_weights_name(self):
        """
        Get all the weights' name

        :return: list of names
        :rtype: list[str]
        """
        return list(self.__weights)

    def is_nodal_weight(self, name):
        """
        Check if a value represent a nodal weight

        :param name: name of the weight
        :type name: str
        :return: True if a value represent a nodal weight, else False
        :rtype: bool
        """
        return self.__types.get(name)

    def plot_weight(self, name, dimension=None, interactive=False, legend_label="", max_value=None, min_value=None,
                    palette='viridis', show_axis=False, title=""):
        """
        Plot the graph with weights on vertices or edges

        :param name: name of the weight
        :type name: str
        :param dimension: dimension of plotted weight. For edges is the linewidth (default 3), for vertices is the nodal size (default 100)
        :type dimension: int | float
        :param interactive: if True, display the plot in interactive mode
        :type interactive: bool
        :param legend_label: label of the colormap
        :type legend_label: str
        :param max_value: maximum value to consider for the colormap, if None it's the actual maximum
        :type max_value: float
        :param min_value: minimum value to consider for the colormap, if None it's the actual minimum
        :type min_value: float
        :param palette: colormap palette
        :type palette: str
        :param show_axis: if True, display the axis of the figure
        :type show_axis: bool
        :param title: title of the plot
        :type title: str
        """
        if self.plotter is None:
            print("Cannot plot graph in dimension greater than 3")
        elif self.__weights.get(name) is None:
            print(f"Warning: weight {name} does not exist")
        elif self.__types.get(name):
            if dimension is None:
                dimension = 100
            self.plotter.plot_vertices_weight(self.__weights[name], interactive, legend_label, max_value, min_value,
                                              palette, show_axis, title, dimension)
        else:
            if dimension is None:
                dimension = 3
            self.plotter.plot_edges_weight(self.__weights[name], dimension, interactive, legend_label, max_value,
                                           min_value, palette, show_axis, title)

    def write_vtk(self, filename):
        """
        Write the WeightedGraph on a vtk file

        :param filename: name of the file (without extension)
        :type filename: str
        """
        # Create nodes
        points = vtk.vtkPoints()
        for p in self.vertices:
            points.InsertNextPoint(p)
        # Create edges
        lines = vtk.vtkCellArray()
        for e in self.edges:
            line = vtk.vtkLine()
            line.GetPointIds().SetId(0, e[0])
            line.GetPointIds().SetId(1, e[1])
            lines.InsertNextCell(line)
        # Create vtkPolyData
        polydata = vtk.vtkPolyData()
        polydata.SetPoints(points)
        polydata.SetLines(lines)
        # Add vertices/edge quantities
        for name in self.get_weights_name():
            quantities = vtk.vtkFloatArray()
            quantities.SetName(name)
            quantities.SetNumberOfComponents(1)
            # Insert values
            for v in self.get_weight(name):
                quantities.InsertNextValue(v)
            # Check if it's a nodal weight
            if self.is_nodal_weight(name):
                polydata.GetPointData().AddArray(quantities)
            else:
                polydata.GetCellData().AddArray(quantities)
        # Write on file
        writer = vtk.vtkPolyDataWriter()
        writer.SetInputData(polydata)
        writer.SetFileName(f"{filename}.vtk")
        writer.Write()

    @classmethod
    def read_vtk(cls, filename):
        """
        Read the WeightedGraph from a vtk file

        :param filename: name of the file (without extension)
        :type filename: str
        :return: WeightedGraph
        :rtype: WeightedGraph
        """
        # Read vtk
        reader = vtk.vtkPolyDataReader()
        reader.SetFileName(f"{filename}.vtk")
        reader.Update()
        # Extract vertices and edges
        polydata = reader.GetOutput()
        vertices = np.array(polydata.GetPoints().GetData())
        edges = np.zeros((polydata.GetNumberOfCells(), 2), dtype=np.int64)
        for i in range(edges.shape[0]):
            edges[i] = [polydata.GetCell(i).GetPointId(0), polydata.GetCell(i).GetPointId(1)]
        # Build weighted graph
        graph = cls(vertices, edges)
        # Add node values
        for i in range(polydata.GetPointData().GetNumberOfArrays()):
            arr = polydata.GetPointData().GetArray(i)
            graph.set_vertices_weight(arr.GetName(), np.array(arr))
        # Add edge values
        for i in range(polydata.GetCellData().GetNumberOfArrays()):
            arr = polydata.GetCellData().GetArray(i)
            graph.set_edges_weight(arr.GetName(), np.array(arr))
        # Return object
        return graph



import numpy as np
import pickle
import random

from .graphProperties import GraphProperties
from .plotter import Plotter2d, Plotter3d


class MetricGraph:
    """
    Class that build a MetricGraph object

    :param vertices: vertices coordinates of the graph
    :type vertices: list | np.ndarray
    :param edges: connectivity matrix of the graph
    :type edges: list | np.ndarray
    """

    def __init__(self, vertices, edges):
        # Save edges and vertices
        self.vertices = np.array(vertices)
        self.edges = np.array(edges)
        self.properties = None
        self.plotter = None
        if self.vertices.shape[1] == 2:
            self.plotter = Plotter2d(self.vertices, self.edges)
        if self.vertices.shape[1] == 3:
            self.plotter = Plotter3d(self.vertices, self.edges)

    def __str__(self):
        return f"Vertices of the graph are:\n {self.vertices}\n" \
               f"Edges of the graph are:\n {self.edges}\n" \
               f"Number of vertices: {self.vertices.shape[0]}\n" \
               f"Number of edges: {self.edges.shape[0]}\n"

    # Public interface of the class

    def get_edges_length(self):
        """
        Compute length of edges between vertices.

        :return: Edge length
        :rtype: np.ndarray
        """
        edges_length = np.zeros(self.edges.shape[0])
        for i, e in enumerate(self.edges):
            x1 = self.vertices[e[0], :]
            x2 = self.vertices[e[1], :]
            edges_length[i] = np.linalg.norm(x1 - x2)
        return edges_length

    def get_graph_laplacian(self, w1=None, w2=None):
        """
        Compute graph laplacian. Consider the edge as undirected. Ignore self loops.

        :param w1: optional weight function, if used L_jj = L_jj * w1(edges_length[i])
        :type w1: function
        :param w2: optional weight function, if used L_jk = L_jk * w2(edges_length[i]), j != k
        :type w2: function
        :return: numpy array containing the graph laplacian
        :rtype: np.ndarray
        """
        # Define matrices
        L = np.zeros((self.vertices.shape[0], self.vertices.shape[0]),
                     dtype=np.int64 if w1 is None and w2 is None else np.float64)
        mask = np.zeros((self.vertices.shape[0], self.vertices.shape[0]), dtype=bool)
        if w1 is None:
            w1 = lambda x: 1
        if w2 is None:
            w2 = lambda x: 1
        # Get edges length
        edges_length = self.get_edges_length()
        # Compute weighted laplacian matrix
        for i, (v1, v2) in enumerate(self.edges):
            if v1 != v2 and not mask[v1, v2]:
                L[v1, v1] += 1 * w1(edges_length[i])
                L[v2, v2] += 1 * w1(edges_length[i])
                L[v1, v2] = -1 * w2(edges_length[i])
                L[v2, v1] = -1 * w2(edges_length[i])
                mask[v1, v2] = True
                mask[v2, v1] = True
        return L

    def get_boundary_mask(self):
        """
        Compute naturally boundary vertices (nodes with at maximum one connection, incoming or outgoing).

        :return: boolean numpy array
        :rtype: np.ndarray[bool]
        """
        n_connections = np.zeros((self.vertices.shape[0]), dtype=int)
        mask = np.zeros((self.vertices.shape[0], self.vertices.shape[0]), dtype=bool)
        for v1, v2 in self.edges:
            if v1 != v2 and not mask[v1, v2]:
                n_connections[v1] += 1
                n_connections[v2] += 1
                mask[v1, v2] = True
                mask[v2, v1] = True
        return n_connections < 2

    def get_boundary_coordinates(self):
        """
        Compute the coordinates of boundary vertices.
        Boundary vertices are nodes with at most one incident edge.

        :return: Array of boundary vertex coordinates
        :rtype: np.ndarray
        """
        mask = self.get_boundary_mask()
        coord = np.zeros((np.sum(mask), self.vertices.shape[1]))
        j = 0
        for i, v in enumerate(self.vertices):
            if mask[i]:
                coord[j] = v
                j += 1
        return coord

    def get_inlets(self):
        """
        Compute inlet vertices (nodes with no incoming connections).

        :return: numpy array of inlet vertices
        :rtype: np.ndarray[int]
        """
        n_connections = np.zeros((self.vertices.shape[0]), dtype=int)
        for v1, v2 in self.edges:
            n_connections[v2] += 1
        return np.where(n_connections == 0)[0]

    def get_outlets(self):
        """
        Compute outlet vertices (nodes with no outgoing connections).

        :return: numpy array of outlet vertices
        :rtype: np.ndarray[int]
        """
        n_connections = np.zeros((self.vertices.shape[0]), dtype=int)
        for v1, v2 in self.edges:
            n_connections[v1] += 1
        return np.where(n_connections == 0)[0]

    def get_properties(self):
        """
        Compute and return properties of the graph.

        :rtype: GraphProperties
        """
        if self.properties is None:
            self.properties = GraphProperties.get_properties_from_graph(self.vertices, self.edges)
        return self.properties

    @classmethod
    def generate_graph(cls, n_vertices, n_dim=3, domain_limits=None, minimum_edges=None, directed=False, loop=False,
                       strong_connected=False, weak_connected=True, acyclic=True):
        """
        Generate MetricGraph object, given its properties.

       :param n_vertices: number of vertices of the graph
       :type n_vertices: int
       :param n_dim: dimension of the space
       :type n_dim: int
       :param domain_limits: limits of the bounding box containing the graph
       :type domain_limits: list | np.ndarray
       :param minimum_edges: number of minimum edges tried to be added (it may happen that the maximum number
                             of possible edges is smaller)
       :type minimum_edges: int
       :param directed: if True directed graph are considered
       :type directed: bool
       :param loop: if True and directed=True self loop are allowed
       :type loop: bool
       :param strong_connected: if True the graph has to be strongly connected (notice that the definition makes sense
                                only for directed graph)
       :type strong_connected: bool
       :param weak_connected: if True the graph has to be weakly connected in case of directed graph or just connected
                              in case of undirected graph
       :type weak_connected: bool
       :param acyclic: if True the graph has to be acyclic
       :type acyclic: bool
       :return: MetricGraph object
       :rtype: MetricGraph
       """
        properties = GraphProperties(n_vertices=n_vertices, n_dim=n_dim, limits=domain_limits,
                                     minimum_edges=minimum_edges, directed=directed,
                                     loop=loop, strong_connected=strong_connected,
                                     weak_connected=weak_connected, acyclic=acyclic)
        # Create graph
        graph = cls(get_random_vertices(properties), get_random_edges(properties))
        # Save graph properties
        graph.properties = properties
        graph.properties.update_n_cycles(graph.vertices, graph.edges)
        # Return graph
        return graph

    def plot(self, color="blue", directed=False, interactive=False, line_widths=2, show_axis=False, title="",
             vertices_label=True):
        """
        Plot image with graph representation.

        :param color: color to use for plotting the edges
        :type color: str
        :param directed: if True, plot the graph using arrows
        :type directed: bool
        :param interactive: if True, display the plot in interactive mode
        :type interactive: bool
        :param line_widths: width of the lines
        :type line_widths: int
        :param show_axis: if True, display the axis of the figure
        :type show_axis: bool
        :param title: title of the plot
        :type title: str
        :param vertices_label: if True, label the vertices of the graph
        :type vertices_label: bool
        """
        if self.plotter is None:
            print("Cannot plot graph in dimension greater than 3")
        else:
            self.plotter.plot_graph(color, directed, interactive, line_widths, show_axis, title, vertices_label)

    def distance(self, point):
        """
        Compute distance between point and the graph --> d(p, G). It returns: \n
        - ret['distance'] = distance between point and the graph, \n
        - ret['space_prj'] = projection on the graph in the nd space, \n
        - ret['abscissa'] = abscissa of the projection, \n
        - ret['nearest_edge'] = nearest edge

        :param point: point to compute distance from
        :type point: list[float] | np.ndarray[float]
        :return: dictionary of computed values
        :rtype: dict
        """
        p = np.array(point)
        # Define points a and b
        a = self.vertices[self.edges[:, 0]]
        b = self.vertices[self.edges[:, 1]]
        # Define segments
        ab = b - a
        ap = p - a
        # Compute projection coefficient
        ab_norm2 = np.sum(ab * ab, axis=1)
        with np.errstate(divide='ignore', invalid='ignore'):
            t = np.sum(ab * ap, axis=1) / ab_norm2
        t = np.where(ab_norm2 == 0.0, 0.0, t)  # for segment with length = 0
        # Limit coefficient on the segment: min(1, max(0, t))
        t_s = np.clip(t, 0, 1)
        # Compute projection
        projection = a + t_s[:, None] * ab
        # Compute distances
        dist = np.linalg.norm(p - projection, axis=1)
        # Get minimum distance
        idx = np.argmin(dist)
        # Return distance, projection, abscissa, and edge index
        edge_len = np.linalg.norm(self.vertices[self.edges[idx, 0]] - self.vertices[self.edges[idx, 1]])
        ret = {
            'distance': dist[idx],
            'space_prj': projection[idx],
            'abscissa': t_s[idx] * edge_len,
            'nearest_edge': idx
        }
        return ret

    def refine_graph(self, n_refinements):
        """
        Refine the graph adding internal vertices, return the refined graph.

        :param n_refinements: number of internal vertices to add for each edge
        :type n_refinements: int
        :return: refined graph
        :rtype: MetricGraph
        """
        new_nodes = self.vertices.tolist()
        new_edges = []
        # For each edge, add the new vertices
        for (i, j) in self.edges:
            new_i = i
            for new_node in np.linspace(self.vertices[i], self.vertices[j], n_refinements + 2)[1:-1]:
                # Insert node
                new_nodes.append(new_node)
                new_j = len(new_nodes) - 1
                # Connect nodes
                new_edges.append([new_i, new_j])
                new_i = new_j
            # Connect last node
            new_edges.append([new_i, j])

        return self.__class__(new_nodes, new_edges)

    def save(self, name):
        """
        Save graph object to file.

        :param name: path and name of the file (without extension)
        :type name: str
        """
        with open(name + '.pkl', 'wb') as out:
            pickle.dump(self, out, pickle.HIGHEST_PROTOCOL)

    @classmethod
    def load(cls, name):
        """
        Read graph object from file.

        :param name: path and name of the file (without extension)
        :type name: str
        :return: Graph read from file
        :rtype: MetricGraph
        """
        with open(name + '.pkl', 'rb') as inp:
            obj = pickle.load(inp)
            if not isinstance(obj, cls):
                return None
            return obj


# Static functions
def get_random_vertices(properties):
    """
    Generate random vertex coordinates given certain properties.
    :param properties: properties of the graph
    :type properties: GraphProperties
    """
    vertices = np.zeros((properties.n_vertices, properties.n_dim))
    for k in range(properties.n_vertices):
        for i in range(properties.n_dim):
            vertices[k][i] = random.uniform(properties.limits[i][0], properties.limits[i][1])
    return vertices


def get_random_edges(properties):
    """
    Generate random edges given certain properties.
    :param properties: properties of the graph
    :type properties: GraphProperties
    """
    # Maximum number of edges addable
    max_edges = properties.get_max_edges()
    # Number of edges to add that avoids cycles
    acyclic_edges = 0
    if properties.weak_connected:
        acyclic_edges = properties.n_vertices - 1
        if properties.strong_connected:
            acyclic_edges += 1
    elif properties.acyclic:
        acyclic_edges = min(properties.n_vertices - 1, properties.minimum_edges)
    # Number of random edges to add
    random_edges = max(min(max_edges, properties.minimum_edges) - acyclic_edges, 0)
    # Initialize and fill edges matrix
    edges = np.zeros(shape=(acyclic_edges + random_edges, 2), dtype=np.int64)
    add_acyclic_edges(edges, acyclic_edges, properties.n_vertices, properties.strong_connected)
    for k in range(random_edges):
        add_random_edge(edges, acyclic_edges + k, properties, properties.directed and properties.acyclic)

    return edges


def add_acyclic_edges(edges, n_edges, n_vertices, strong_connect=False):
    """
    Add n_edges to the graph in such a way we don't have cycles.
    Note that the n_edges must be the first edges added in the graph.
    :param edges: edges of the graph
    :type edges: np.ndarray
    :param n_edges: number of edges to add
    :type n_edges: int
    :param n_vertices: number of vertices of the graph
    :type n_vertices: int
    :param strong_connect: if True then it creates a strong connected graph with one cycle
    """
    # In case of strong connected graph, we add the last edges after the loop
    if strong_connect:
        n_edges -= 1
    unselected_v = [i for i in range(n_vertices)]
    i = random.randrange(0, len(unselected_v))
    selected_v = [unselected_v[i]]
    unselected_v.pop(i)
    for k in range(n_edges):
        i = random.randrange(0, len(unselected_v))
        if strong_connect:
            # Take the index of the last element connected to the graph, since we will connect the new vertex to it
            j = len(selected_v) - 1
        else:
            j = random.randrange(0, len(selected_v))
        # Add edge
        edges[k] = [unselected_v[i], selected_v[j]]
        selected_v.append(unselected_v[i])
        unselected_v.pop(i)
    # Add last edge
    if strong_connect:
        edges[n_edges] = [selected_v[0], selected_v[-1]]


def add_random_edge(edges, idx, properties, keep_acyclicity=False):
    """
    Add a random edge to the graph.
    :param edges: edges of the graph
    :type edges: np.ndarray
    :param idx: index of the row of the edge matrix where to add the edge
    :type idx: int
    :param properties: properties of the graph
    :type properties: GraphProperties
    :param keep_acyclicity: if True add an edge keeping the graph acyclic
    """
    while True:
        a = random.randrange(0, properties.n_vertices)
        while True:
            b = random.randrange(0, properties.n_vertices)
            if properties.directed and properties.loop:
                break
            if a != b:
                break
        if not is_edge_present(edges, [a, b], properties.directed):
            if keep_acyclicity and is_edge_present(edges, [b, a], properties.directed):
                continue
            edges[idx] = [a, b]
            if keep_acyclicity and has_dgraph_cycle(edges, properties.n_vertices):
                edges[idx] = [b, a]
            break


def is_edge_present(edges, e, directed):
    """
    Verify if an edge is present in the graph.
    :param edges: edges of the graph
    :type edges: np.ndarray
    :param e: edge to verify in the form [a,b]
    :type e: list
    :param directed: if True the edge are directed
    :type directed: bool
    :return: True if the edge is present
    """
    if directed:
        return any(np.equal(edges, e).all(1))
    return any(np.equal(edges, e).all(1)) or any(np.equal(edges, [e[1], e[0]]).all(1))


def has_dgraph_cycle(edges, n_vertices):
    """
    Implement a topological sort to verify if the graph has a cycle.
    :param edges: edges of the graph
    :type edges: np.ndarray
    :param n_vertices: number of vertices of the graph
    :type n_vertices: int
    :return: True if directed graph has at least a cycle
    """
    # Create a copy of list self.edges
    graph = edges.tolist()

    def has_no_incoming_edge(node):
        for e in graph:
            if e[1] == node:
                return False
        return True

    # Initialize empty list L
    L = []
    # Create set of nodes with no incoming edges
    Q = []
    for i in range(n_vertices):
        if has_no_incoming_edge(i):
            Q.append(i)
    # While Q is not empty
    while len(Q) > 0:
        n = Q.pop(0)
        L.append(n)
        # For each node m with an edge e from n to m
        for m in range(n_vertices):
            if [n, m] in graph:
                # Remove edge e from the graph
                graph.remove([n, m])
                # If m has no other incoming edges
                if has_no_incoming_edge(m):
                    Q.append(m)
    if len(graph) > 0:
        return True
    # Proposed topologically sorted order: L
    return False



import numpy as np


class GraphProperties:
    """
    Class that contains the properties of the graph in Graph object.
    """

    def __init__(self, n_vertices, n_dim=3, limits=None, minimum_edges=None, directed=True, loop=False,
                 strong_connected=False, weak_connected=True, acyclic=True):
        """
        :param n_vertices: number of vertices of the graph
        :type n_vertices: int
        :param n_dim: dimension of the space
        :param limits: limits of the bounding box containing the graph
        :type limits: list | np.ndarray
        :param minimum_edges: number of minimum edges tried to be added (it may happen that the maximum number
                              of possible edges is smaller). Default is n_vertices - 1.
        :type minimum_edges: int
        :param directed: if True directed graph are considered
        :param loop: if True and directed=True self loop are allowed
        :param strong_connected: if True the graph has to be strongly connected (notice that the definition makes sense
                                 only for directed graph)
        :param weak_connected: if True the graph has to be weakly connected in case of directed graph or just connected
                               in case of undirected graph
        :param acyclic: if True the graph has to be acyclic
        """
        # Number of vertices
        self.n_vertices = int(n_vertices)
        if self.n_vertices < 1:
            self.n_vertices = 1
        # Number of edges
        if minimum_edges is None:
            self.minimum_edges = self.n_vertices - 1
        else:
            self.minimum_edges = int(minimum_edges)
        # Dimension of the space
        self.n_dim = int(n_dim)
        if self.n_dim < 2:
            self.n_dim = 2
        # Limits of the domain
        if limits is None:
            self.limits = np.zeros((self.n_dim, 2))
            self.limits[:, 0], self.limits[:, 1] = -1, 1
        else:
            self.limits = np.array(limits)

        self.directed = directed
        self.loop = loop
        self.strong_connected = strong_connected
        self.weak_connected = weak_connected
        self.acyclic = acyclic
        if self.acyclic:
            # We can't have loop in acyclic graph
            self.loop = False
        if not self.directed:
            # For undirected graph strong connection doesn't make sense:
            self.strong_connected = False
        if self.strong_connected:
            # Strong connection implies weak connection:
            self.weak_connected = True
            # Strong connection implies cyclic graph:
            self.acyclic = False
        # Number of cycle in the graph
        self.n_cycles = -1

    def __str__(self):
        return f"Dimension of the space: {self.n_dim}\n" \
               f"Limits of the bounding box are \n {self.limits}\n" \
               f"Is the graph directed: {self.directed}\n" \
               f"Does the graph allow self loop: {self.loop}\n" \
               f"Has the graph to be strongly connected: {self.strong_connected}\n" \
               f"Has the graph to be weakly connected: {self.weak_connected}\n" \
               f"Has the graph to be acyclic: {self.acyclic}\n" \
               f"Number of cycles in the graph: {self.n_cycles}"


    # Public interface

    def get_max_edges(self):
        """
        Compute the maximum edges the graph can have given its topology
        :return: maximum number of edges
        """
        max_edges = self.n_vertices ** 2
        if self.directed:
            if self.acyclic:
                max_edges = (max_edges - self.n_vertices) / 2
            elif not self.loop:
                max_edges -= self.n_vertices
        else:
            if self.acyclic:
                max_edges = self.n_vertices - 1
            else:
                max_edges = (max_edges - self.n_vertices) / 2
        return int(max_edges)

    def update_n_cycles(self, vertices, edges):
        """
        Update the number of cycles of the graph (after random generate the graph)
        :param vertices: vertices of the graph of shape (n_vertices, n_dim)
        :type vertices: np.ndarray
        :param edges: edges of the graph of shape (n_edges, 2)
        :type edges: np.ndarray
        """
        if self.acyclic:
            self.n_cycles = 0
        else:
            if self.directed:
                self.n_cycles = count_d_cycles(vertices, edges)
            else:
                edges = duplicate_edges(edges)
                self.n_cycles = count_nd_cycles(vertices, edges)

    @classmethod
    def get_properties_from_graph(cls, vertices, edges):
        """
        Compute graph properties given vertices and edges.
        Notice that to be an undirected graph, for each (i,j) in edges also (j,i) must be present
        :param vertices: vertices of the graph of shape (n_vertices, n_dim)
        :type vertices: list | np.ndarray
        :param edges: edges of the graph of shape (n_edges, 2)
        :type edges: list | np.ndarray
        :return: GraphProperties instance
        """
        vertices = np.array(vertices)
        edges = np.array(edges)
        # Find graph properties
        weak_connected, strong_connected = False, False
        self_loop = False
        directed = check_directed(edges)
        if directed:
            weak_connected, strong_connected = check_d_connected(vertices, edges)
            n_self_loop = count_self_loop(edges)
            self_loop = n_self_loop > 0
            n_cycles = count_d_cycles(vertices, edges)
        else:
            weak_connected = check_nd_connected(vertices, edges)
            n_cycles = count_nd_cycles(vertices, edges)

        # Create GraphProperties object
        p = cls(n_vertices=vertices.shape[0],
                n_dim=vertices.shape[1],
                limits=get_bounding_box(vertices),
                directed=directed,
                loop=self_loop,
                strong_connected=strong_connected,
                weak_connected=weak_connected,
                acyclic=n_cycles == 0)
        # Add number of cycles
        p.n_cycles = n_cycles
        # Return object
        return p

# Static functions

def get_bounding_box(vertices):
    """
    Compute bounding box of the graph
    :param vertices: vertices of the graph
    :type vertices: np.ndarray
    :return: numpy array of shape (n_dim, 2)
    """
    limits = np.zeros((vertices.shape[1], 2))
    for i in range(vertices.shape[1]):
        limits[i, 0] = min(vertices[:, i])
        limits[i, 1] = max(vertices[:, i])
    return limits


def check_directed(edges):
    """
    Check if the graph is directed
    :param edges: edges of the graph
    :type edges: np.ndarray
    :return: True if the graph is directed, False otherwise
    """
    if count_self_loop(edges) != 0:
        return True
    for e in edges:
        if not any(np.equal(edges, [e[1], e[0]]).all(1)):
            return True
    return False


def check_nd_connected(vertices, edges):
    """
    Check if an undirected graph is connected
    :param vertices: vertices of the graph
    :type vertices: np.ndarray
    :param edges: edges of the graph
    :type edges: np.ndarray
    :return: True if the graph is directed, False otherwise
    """
    n_vert = vertices.shape[0]
    is_connected = np.zeros(n_vert)
    if n_vert - 1 > edges.shape[0]:
        return False
    for i in range(n_vert):
        for e in edges:
            if e[0] == i:
                is_connected[e[1]] += 1
    for i in is_connected:
        if i == 0:
            return False
    return True


def check_d_connected(vertices, edges):
    """
    Check if a directed graph is weak connected and strong connected
    :param vertices: vertices of the graph
    :type vertices: np.ndarray
    :param edges: edges of the graph
    :type edges: np.ndarray
    :return: first bool for weak connected, second bool for strong connected
    """
    # Remove self loops
    ed = []
    for e in edges:
        if e[0] != e[1]:
            ed.append(e)
    ed = np.array(ed)
    # Check strong connection
    is_strong_connected = check_nd_connected(vertices, ed)
    if is_strong_connected:
        return True, True
    # Check weak connection
    nd_edges = duplicate_edges(ed)
    is_weak_connected = check_nd_connected(vertices, nd_edges)
    if is_weak_connected:
        return True, False
    return False, False


def duplicate_edges(edges):
    """
    It ensures that if (i,j) belongs to edges, also (j,i) will be present
    :param edges: edges of the graph
    :type edges: np.ndarray
    :return: numpy array with duplicate edges
    """
    add = []
    for e in edges:
        if not any(np.equal(edges, [e[1], e[0]]).all(1)):
            add.append([e[1], e[0]])
    if not add:
        return edges
    add = np.array(add)
    return np.vstack((edges, add))


def count_self_loop(edges):
    """
    Count the number of self loops in the graph
    :param edges: edges of the graph
    :type edges: np.ndarray
    :return: number of self loops
    """
    count = 0
    for e in edges:
        if e[0] == e[1]:
            count += 1
    return count


def count_simple_cycles(vertices, edges):
    """
    Count the number of simple cycles in the graph
    :param vertices: vertices of the graph
    :type vertices: np.ndarray
    :param edges: edges of the graph
    :type edges: np.ndarray
    :return: number of simple cycles
    """
    n_vert = vertices.shape[0]
    visited = [False] * (n_vert + 1)
    stack = []
    edges = convert_edges(n_vert, edges)
    count = 0

    for v in range(0, n_vert):
        count = dfs(v, visited, stack, edges, count)
        visited[v] = True

    return count


def convert_edges(n_vert, edges):
    """
    Change representation of edges matrix
    :param n_vert: number of vertices in the graph
    :type n_vert: int
    :param edges: edges of the graph
    :type edges: np.ndarray
    :return: new representation
    """
    ed = [[] for i in range(n_vert)]
    for e in edges:
        if e[0] != e[1]:
            ed[e[0]].append(e[1])
    return ed


def dfs(vertex, visited, stack, edges, count):
    """
    Implement depth-first search algorithm
    :param vertex: root vertex
    :type vertex: int
    :param visited: list of visited vertices
    :type visited: list
    :param stack: stack used
    :type stack: list
    :param edges: alternative representation of edges matrix
    :type edges: list[list]
    :param count: current number of cycles found
    :type count: int
    :return: number of cycles found
    """
    visited[vertex] = True
    stack.append(vertex)

    for neighbor in edges[vertex]:
        if not visited[neighbor]:
            count = dfs(neighbor, visited, stack, edges, count)
        elif neighbor in stack:
            if neighbor != stack[-2] and neighbor == stack[0]:
                # print("Cycle found:", stack + [neighbor])
                count += 1

    stack.pop()
    visited[vertex] = False
    return count


def count_d_cycles(vertices, edges):
    """
    Count the total number of cycles in a directed graph (including self loop and two edges cycles)
    :param vertices: vertices of the graph
    :type vertices: np.ndarray
    :param edges: edges of the graph
    :type edges: np.ndarray
    :return: number of cycles
    """
    simple_cycles = count_simple_cycles(vertices, edges)
    self_loops = count_self_loop(edges)
    two_cycles = 0
    for e in edges:
        if any(np.equal(edges, [e[1], e[0]]).all(1)):
            two_cycles += 1
    two_cycles = (two_cycles - self_loops) // 2

    return simple_cycles + self_loops + two_cycles


def count_nd_cycles(vertices, edges):
    """
    Count the total number of cycles in an undirected graph
    :param vertices: vertices of the graph
    :type vertices: np.ndarray
    :param edges: edges of the graph
    :type edges: np.ndarray
    :return: number of cycles
    """
    return count_simple_cycles(vertices, edges) // 2



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'metricgraph'
copyright = '2025, ploki99'
author = 'ploki99'
release = '0.1.1'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx_autodoc_typehints",
]

templates_path = ['_templates']
exclude_patterns = []



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']

# -- Add project path --------------------------------------------------------
import os
import sys
sys.path.insert(0, os.path.abspath("../.."))